from typing import List

import swem
from gensim.models import KeyedVectors


def tokenize_en(text: str) -> List[str]:
    text_processed = text.replace('.', ' .').replace(',', ' ,')
    return text_processed.replace('?', ' ?').replace('!', ' !').split()


if __name__ == '__main__':
    kv = KeyedVectors.load('wiki_mecab-ipadic-neologd.kv')
    swem_embed = swem.SWEM(kv, tokenizer=tokenize_en)

    doc: str = 'This is an implementation of SWEM.'
    embed = swem_embed.infer_vector(doc, method='max')
    print(embed)
    print(embed.shape)



from typing import List

import MeCab
import swem
from gensim.models import KeyedVectors


def tokenize_ja(text: str, args: str = '-O wakati') -> List[str]:
    tagger = MeCab.Tagger(args)
    return tagger.parse(text).strip().split(' ')


if __name__ == '__main__':
    kv = KeyedVectors.load('wiki_mecab-ipadic-neologd.kv')
    swem_embed = swem.SWEM(kv, tokenize_ja)

    doc: str = 'すもももももももものうち'
    embed = swem_embed.infer_vector(doc, method='max')
    print(embed)
    print(embed.shape)



from typing import Callable, List, Tuple

from gensim.models.keyedvectors import Word2VecKeyedVectors
import numpy as np


def _word_embed(
    token: str,
    kv: Word2VecKeyedVectors,
    uniform_range: Tuple[float, ...] = (-0.01, 0.01)
) -> np.ndarray:
    """ Get word embedding of given token.

    Args:
        token (str): A word token to get embed.
        kv (Word2VecKeyedVectors): Vocabularies dictionary.
        uniform_range (Tuple[float, ...]): A range of uniform distribution to
                                           generate random vector.

    Returns:
        numpy.ndarray: An array with shape (self.embed_dim, )
    """
    try:
        return kv[token]
    except Exception as e:
        print(e)
        embed_dim: int = kv.vector_size
        return np.random.uniform(
            uniform_range[0],
            uniform_range[1],
            embed_dim
        )


def _word_embeds(tokens: List[str], kv: Word2VecKeyedVectors,
                 uniform_range: Tuple[float, ...]) -> np.ndarray:
    """ Get word embeddings of given tokens.

    Args:
        tokens (List[str]): A word tokens to calculate embeddding.

    Returns:
        numpy.ndarray: An embedding array with shape
                       (token_size, self.embed_dim, ).
    """
    doc_embed: List[np.ndarray] = []
    for token in tokens:
        word_embed: np.ndarray = _word_embed(
            token=token, kv=kv, uniform_range=uniform_range
        )
        doc_embed.append(word_embed)
    return np.array(doc_embed)


def _hierarchical_pool(
    tokens_embed: np.ndarray,
    num_windows: int = 3
) -> np.ndarray:
    """ Hierarchical Pooling: It takes word-order or spatial information
        into consideration when calculate document embeddings.

    Args:
        tokens_embed (np.ndarray): An embeded document vector.
        num_windows (int): A sizw of window to consider sequence.

    Returns:
        numpy.ndarray: An embedding array with shape (self.embed_dim, ).
    """
    text_len: int = tokens_embed.shape[0]
    if num_windows > text_len:
        raise ValueError(f'window size [{num_windows}] must be less '
                         f'than text length{text_len}.')

    num_iters: int = text_len - num_windows + 1
    pooled_doc_embed: List[np.ndarray] = [
        np.mean(tokens_embed[i:i + num_windows],
                axis=0) for i in range(num_iters)
    ]
    return np.max(pooled_doc_embed, axis=0)


def infer_vector(
    tokens: List[str],
    kv,
    method: str = 'avg',
    uniform_range: Tuple[float, float] = (-0.01, 0.01),
    num_windows: int = 3
) -> np.ndarray:
    tokens_embed: np.ndarray = _word_embeds(
        tokens=tokens,
        kv=kv,
        uniform_range=uniform_range
    )

    if method == 'max':
        return tokens_embed.max(axis=0)

    elif method == 'avg':
        return tokens_embed.mean(axis=0)

    elif method == 'concat':
        return np.hstack([tokens_embed.mean(axis=0), tokens_embed.max(axis=0)])

    elif method == 'hierarchical':
        return _hierarchical_pool(tokens_embed, num_windows)

    else:
        raise ValueError(
            f'infer_vector has no attribute [{method}] method.'
        )


class SWEM:
    """Implementation of SWEM.

    Args:
        kv: gensim.models.keyedvectors.Word2VecKeyedVectors
        tokenizer: Callable
            Callable object to tokenize input text.
        uniform_range: Tuple[float, ...]
            A range of uniform distribution to create random embedding.
    """

    def __init__(self, kv, tokenizer: Callable, uniform_range=(-0.01, 0.01)):
        self.kv: Word2VecKeyedVectors = kv

        if not callable(tokenizer):
            raise ValueError('tokenizer must be callable object.')
        self.tokenizer: Callable = tokenizer
        self.uniform_range: Tuple[float, ...] = uniform_range

    def infer_vector(
        self,
        doc: str,
        method: str = 'max',
        num_windows: int = 3
    ) -> np.ndarray:
        """ A main method to get document vector.

        Args:
            doc (str): A document str to get embeddings.
            method (str): Designate method to pool.
                         ('max', 'avg', 'concat', 'hierarchical')

        Returns:
            numpy.ndarray: An embedding array.
        """
        tokens: List[str] = self.tokenizer(doc)
        doc_embed: np.ndarray = _word_embeds(
            tokens=tokens,
            kv=self.kv,
            uniform_range=self.uniform_range
        )

        if method == 'max':
            return doc_embed.max(axis=0)

        elif method == 'avg':
            return doc_embed.mean(axis=0)

        elif method == 'concat':
            return np.hstack([doc_embed.mean(axis=0), doc_embed.max(axis=0)])

        elif method == 'hierarchical':
            return _hierarchical_pool(doc_embed, num_windows)

        else:
            raise ValueError(
                f'infer_vector has no attribute [{method}] method.'
            )



from __future__ import absolute_import
from __future__ import unicode_literals
import os

from setuptools import setup, find_packages


def get_long_description():
    readme_filepath = os.path.join(os.path.dirname(__file__), 'README.md')
    with open(readme_filepath) as f:
        return f.read()


def get_version():
    version_filepath = os.path.join(os.path.dirname(__file__), 'swem', 'version.py')
    with open(version_filepath) as f:
        for line in f:
            if line.startswith('__version__'):
                return line.strip().split()[-1][1:-1]


def get_install_requires():
    requirements_filepath = os.path.join(os.path.dirname(__file__), 'requirements.txt')
    with open(requirements_filepath, 'r') as f:
        return f.read().splitlines()


def get_tests_requires():
    requirements_filepath = os.path.join(os.path.dirname(__file__), 'requirements-dev.txt')
    with open(requirements_filepath, 'r') as f:
        return f.read().splitlines()


setup(
    name='swem',
    version=get_version(),
    url='https://github.com/yutayamazaki/swem',
    author='Yuta Yamazaki',
    author_email='yu.yamazakii@gmail.com',
    maintainer='Yuta Yamazaki',
    maintainer_email='yu.yamazakii@gmail.com',
    description='A portable document embedding using SWEM.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=get_install_requires(),
    tests_require=get_tests_requires(),
    license='MIT',
    keywords='swem nlp python',
)



import unittest
from typing import Dict, List, Tuple

import MeCab
import numpy as np
import pytest

import swem
from swem import models


def tokenize_ja(text: str, args: str = '-O wakati') -> List[str]:
    tagger = MeCab.Tagger(args)
    return tagger.parse(text).strip().split(' ')


def test_word_embed():
    token = '私'
    kv = MockKV()
    embed = models._word_embed(token, kv=kv)
    assert embed.shape == (200, )


def test_word_embeds():
    tokens = ['私', 'は']
    kv = MockKV()
    embed = models._word_embeds(tokens, kv=kv, uniform_range=(-0.01, 0.01))
    assert embed.shape == (2, 200)


def test_hierarchical_pool():
    text: str = 'すもももももももものうち'
    kv: MockKV = MockKV()
    word_embeds: np.ndarray = models._word_embeds(text, kv, (-1, 1))
    ret: np.ndarray = models._hierarchical_pool(word_embeds, num_windows=3)
    assert ret.shape == (200, )


def test_hierarchical_pool_raise():
    """ ValueError: when invalid n_windows passed. """
    doc: str = '桃'
    kv: MockKV = MockKV()
    word_embeds = models._word_embeds(doc, kv, (-1, 1))
    with pytest.raises(ValueError):
        # text_length: 1, n_windows: 3
        models._hierarchical_pool(word_embeds, num_windows=3)


def test_infer_vector_functional():
    tokens = ['私', 'は', '私', 'は']
    kv = MockKV()
    methods: Dict[str, Tuple[int]] = {
        'avg': (200, ),
        'max': (200, ),
        'concat': (400, ),
        'hierarchical': (200, )
    }
    for method, shape in methods.items():
        embed: np.ndarray = swem.infer_vector(tokens, kv=kv, method=method)
        assert embed.shape == shape


class MockKV:

    vector_size = 200

    def __getitem__(self, item):
        return np.zeros(self.vector_size)


class SWEMTests(unittest.TestCase):

    def setUp(self):
        self.swem = models.SWEM(MockKV(), tokenize_ja)

    def init_tokenizer_is_not_callable(self):
        with pytest.raises(ValueError):
            models.SWEM(MockKV(), 'aaaa')

    def test_infer_vector(self):
        methods = {
            'avg': 200,
            'concat': 400,
            'hierarchical': 200,
            'max': 200,
        }
        doc = 'すもももももももものうち'
        for method_name, embed_dim in methods.items():
            ret = self.swem.infer_vector(doc, method=method_name)
            assert ret.shape == (embed_dim, )

    def test_infer_vector_raise(self):
        doc = 'すもももももももものうち'
        method = 'invalid method'
        with pytest.raises(ValueError):
            self.swem.infer_vector(doc, method=method)
